define([
    'require',
    'lodash',
    'plotly.js/dist/plotly.min'
], function (require, _, Plotly) {
    'use strict';

    window.onresize = function () {
        try {
            resize();
        } catch (error) {
            console.info(error);
        }
    };

    function resize() {
        var d3 = Plotly.d3;
        var gd3 = $('.chart')
            .each(function (i, d) {
                // bypass hidden area which resize can't work
                if ($(d).is(":hidden")) {
                    return true;
                }
                var gd = d3.select(d).node();
                if (gd && gd.data) {
                    Plotly.Plots.resize(gd)
                };
            });
    }

    var _param = function (para, metricName) {
        return para === metricName ? 'best_metric' : ('param_' + para)
    }

    var _paramLabel = function (para, metricName) {
        return (para === metricName ? 'metric- ' : 'param- ') + para;
    }

    var chartModule = {
        // Function to render chart
        lineChart: function PlotLineChart(elementRef, chartData, showLegend, dTick, xtype, sortBy) {

            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };

            var top5;
            if (showLegend && sortBy) {
                chartData.series.sort((x, y) => {
                    return parseInt(x.run_id) - parseInt(y.run_id);
                });

                top5 = chartData.series.slice(0).sort((x, y) => {
                    return sortBy.order === 'asc' ?
                        (Math.min.apply(Math, y.data) - Math.min.apply(Math, x.data)) : (Math.max.apply(Math, y.data) - Math.max.apply(Math, x.data));
                }).map(r=>r.run_id).slice(0,5);
            }
            var metricData = chartData.series;
            // Rename data to y and add type as preferred by Plotly
            for (var i = 0; i < chartData.series.length; i++) {
                if (showLegend && top5 && top5.indexOf(parseInt(chartData.series[i].run_id)) === -1) {
                    metricData[i]['visible'] = 'legendonly';
                }
                metricData[i]['y'] = chartData.series[i]['data'];
                metricData[i]['x'] = chartData.series[i].categories ? chartData.series[i].categories : chartData.categories;
                if (chartData.series[i]['stepped'])
                    metricData[i]['line'] = {
                        shape: 'hv'
                    };
                delete chartData.series[i]['data'];
            };

            var chartMargin = {
                l: 40,
                r: 0,
                b: 40,
                t: 40,
                pad: 4
            }

            var layout = {
                autosize: true,
                title: chartData.title,
                margin: chartMargin,
                showlegend: showLegend,
                xaxis: {}
            }

            if (dTick) {
                layout.xaxis.dtick = dTick;
            }

            if (xtype) {
                layout.xaxis.type = xtype
            }

           Plotly.react(elementRef, metricData, layout, config);
        },
        multiPrecisionRecall: function PlotMultiPrecisionRecall(elementRef, precisionRecallMetrics) {
            var chartData = _.chain(precisionRecallMetrics)
                .keys()
                .filter(function (key) {
                    return _.startsWith(key, 'recall_');
                })
                .map(function (key) {
                    var className = (key.replace('recall_', ''));

                    var trace = _.reduce(precisionRecallMetrics[key], function (result, xVal, idx) {
                        var yVal = precisionRecallMetrics['precision_' + className][idx];
                        if (xVal === 'NaN' || yVal === 'NaN') {
                            return result;
                        }
                        result.x.push(xVal);
                        result.y.push(yVal);
                        return result;
                    }, {
                        x: [],
                        y: []
                    });
                    trace.mode = 'lines+markers';
                    if (className === 'macro') {
                        trace.name = 'Macro Average';
                    } else if (className === 'micro') {
                        trace.name = 'Micro Average';
                    } else {
                        trace.name = 'Class ' + _.replace(className, 'class_', '');
                        trace.mode = 'lines';
                        trace.visible = 'legendonly';
                        trace.line = {
                            dash: 'dot',
                            width: 4
                        };
                    }
                    return trace;
                }).value();

            var idealTrace = {
                x: [0, 1, 1],
                y: [1, 1, 0],
                mode: 'lines',
                name: 'Ideal',
                line: {
                    dash: 'dashdot',
                    width: 4,
                    color: "rgb(49, 150, 63)"
                }
            };
            chartData.push(idealTrace);

            var layout = {
                title: 'Multi-class Precision-Recall',
                xaxis: {
                    title: 'Recall',
                    showgrid: true,
                    zeroline: true
                },
                yaxis: {
                    title: 'Precision',
                    showline: false
                },
                hovermode: 'closest',
                autosize: true
            };
            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };
            Plotly.react(elementRef, chartData, layout, config);
        },

        ROCCurve: function PlotROCCurve(elementRef, rocData) {
            var chartData = _.chain(rocData)
                .keys()
                .filter(function (key) {
                    return _.startsWith(key, 'false_positive_rate_');
                })
                .map(function (key) {
                    var className = (key.replace('false_positive_rate_', ''));

                    var trace = _.reduce(rocData[key], function (result, xVal, idx) {
                        var yVal = rocData['true_positive_rate_' + className][idx];
                        if (xVal === 'NaN' || yVal === 'NaN') {
                            return result;
                        }
                        result.x.push(xVal);
                        result.y.push(yVal);
                        return result;
                    }, {
                        x: [],
                        y: []
                    });
                    trace.mode = 'lines+markers';

                    if (className === "macro") {
                        trace.name = 'Macro Average';
                    } else if (className === "micro") {
                        trace.name = 'Micro Average';
                    } else {
                        trace.name = "Class " + _.replace(className, 'class_', '');
                        trace.mode = 'lines';
                        trace.visible = 'legendonly';
                        trace.line = {
                            dash: 'dot',
                            width: 4
                        };
                    }

                    return trace;
                }).value();

            var idealTrace = {
                x: [0, 0, 1],
                y: [0, 1, 1],
                mode: 'lines',
                name: 'Ideal',
                line: {
                    dash: 'dashdot',
                    width: 4,
                    color: "rgb(49, 150, 63)"
                }
            };
            chartData.push(idealTrace);

            var layout = {
                title: 'ROC',
                xaxis: {
                    title: 'False Positive Rate',
                    showgrid: true,
                    zeroline: true
                },
                yaxis: {
                    title: 'True Positive Rate',
                    showline: false
                },

                autosize: true
            };
            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };
            Plotly.react(elementRef, chartData, layout, config);
        },

        confusionMatrix: function PlotConfusionMatrix(elementRef, confusion_matrices, labels) {
            var confusionMatrix = confusion_matrices.confusion_matrix;
            var normalizedConfusionMatrix = confusion_matrices.normalized_confusion_matrix;

            var xList = [];
            var yList = [];
            var colorList = [];
            var sizeList = [];
            var dim = Math.sqrt(normalizedConfusionMatrix.length);
            var dotScale = 300 / dim;
            var minimumDotSize = 5;
            var zeroDotSize = 1;
            var barColor = 'rgb(192,192,192)'; // grey
            var axisLabelSize = 16;
            var bgndColor = 'rgb(255,255,255)'; // white

            for (var j = dim; j >= 1; j--) {
                for (var i = 1; i <= dim; i++) {
                    xList.push(i - 1);
                    yList.push(j - 1);

                    // index to access element in confusion matrix array using x and y coordinates
                    var index = (dim - j) * dim + i - 1;

                    // green and red gradients depending on extent of accuracy / inaccuracy
                    // color scale can be adjusted
                    var green =
                        'rgb(0,' +
                        (50 + (1 - normalizedConfusionMatrix[index]) * 255) +
                        ',' +
                        (1 - normalizedConfusionMatrix[index]) * 130 +
                        ')';
                    var red =
                        'rgb(255,' +
                        (1 - normalizedConfusionMatrix[index]) * 204 +
                        ',' +
                        (1 - normalizedConfusionMatrix[index]) * 204 +
                        ')';

                    // green circles along the diagonal, red off the diagonal
                    colorList.push(i + j === dim + 1 ? green : red);

                    if (normalizedConfusionMatrix[index] === 0) {
                        sizeList.push(zeroDotSize);
                        // if value is 0, green circles off the diagonal, red on the diagonal
                        colorList[colorList.length - 1] = colorList[colorList.length - 1] === green ? red : green;
                    } else {
                        // scale circles according to relative size; if size is less than minimum size, set to min
                        var dotSize = normalizedConfusionMatrix[index] * dotScale;
                        sizeList.push(dotSize < minimumDotSize ? minimumDotSize : dotSize);
                    }
                }
            }

            var normalizedTextList = [];
            var unnormalizedTextList = [];
            var vertXAxis = [];
            var horizYAxis = [];
            for (var k = 0; k < yList.length; k++) {
                normalizedTextList.push(normalizedConfusionMatrix[k].toString());
                unnormalizedTextList.push(confusionMatrix[k].toString());

                if (k % dim === 0) {
                    horizYAxis.push(yList[k]);
                }

                if (k < dim) {
                    vertXAxis.push(xList[k]);
                }
            }

            // generate row and column sums for bar graphs
            var normalizedVertList = Array.from(Array(dim), () => 0);
            var unnormalizedVertList = Array.from(Array(dim), () => 0);
            var normalizedHorizList = Array.from(Array(dim), () => 0);
            var unnormalizedHorizList = Array.from(Array(dim), () => 0);

            for (var m = 0; m < normalizedConfusionMatrix.length; m++) {
                var normalizedAddAmt = normalizedConfusionMatrix[m];
                var unnormalizedaddAmt = confusionMatrix[m];

                normalizedVertList[m % dim] = normalizedVertList[m % dim] + normalizedAddAmt;
                unnormalizedVertList[m % dim] = unnormalizedVertList[m % dim] + unnormalizedaddAmt;

                normalizedHorizList[Math.floor(m / dim)] = normalizedHorizList[Math.floor(m / dim)] + normalizedAddAmt;
                unnormalizedHorizList[Math.floor(m / dim)] =
                    unnormalizedHorizList[Math.floor(m / dim)] + unnormalizedaddAmt;
            }

            // confusion matrix
            var normalizedTrace1 = {
                x: xList,
                y: yList,
                marker: {
                    color: colorList,
                    size: sizeList
                },
                mode: 'markers',
                text: normalizedTextList,
                textposition: 'top center',
                type: 'scatter',
                hoverinfo: 'text',
                xaxis: 'x',
                yaxis: 'y',
                visible: false
            };

            var unnormalizedTrace1 = {
                x: xList,
                y: yList,
                marker: {
                    color: colorList,
                    size: sizeList
                },
                mode: 'markers',
                text: unnormalizedTextList,
                textposition: 'top center',
                type: 'scatter',
                hoverinfo: 'text',
                xaxis: 'x',
                yaxis: 'y',
                visible: true
            };

            // column sums bar graph
            var normalizedTrace2 = {
                x: vertXAxis,
                y: normalizedVertList,
                type: 'bar',
                marker: {
                    color: barColor
                },
                xaxis: 'x',
                yaxis: 'y4',
                hoverinfo: 'none',
                visible: false
            };

            var unnormalizedTrace2 = {
                x: vertXAxis,
                y: unnormalizedVertList,
                type: 'bar',
                marker: {
                    color: barColor
                },
                xaxis: 'x',
                yaxis: 'y4',
                hoverinfo: 'none',
                visible: true
            };

            // row sums bar graph
            var normalizedTrace3 = {
                x: normalizedHorizList,
                y: horizYAxis,
                type: 'bar',
                orientation: 'h',
                marker: {
                    color: barColor
                },
                xaxis: 'x2',
                yaxis: 'y',
                hoverinfo: 'none',
                visible: false
            };

            var unnormalizedTrace3 = {
                x: unnormalizedHorizList,
                y: horizYAxis,
                type: 'bar',
                orientation: 'h',
                marker: {
                    color: barColor
                },
                xaxis: 'x2',
                yaxis: 'y',
                hoverinfo: 'none',
                visible: true
            };

            // row class labels
            var trace4 = {
                x: Array.apply(null, Array(dim)).map(Number.prototype.valueOf, 1),
                y: horizYAxis,
                type: 'bar',
                orientation: 'h',
                mode: 'text+markers',
                marker: {
                    color: bgndColor
                },
                text: labels,
                textposition: 'inside',
                xaxis: 'x3',
                yaxis: 'y',
                hoverinfo: 'none',
                visible: true
            };

            // column class labels
            var trace5 = {
                x: vertXAxis,
                y: Array.apply(null, Array(dim)).map(Number.prototype.valueOf, 1),
                type: 'bar',
                mode: 'text+markers',
                marker: {
                    color: bgndColor
                },
                text: labels,
                textposition: 'inside',
                xaxis: 'x',
                yaxis: 'y3',
                hoverinfo: 'none',
                visible: true
            };

            var data = [
                normalizedTrace1,
                unnormalizedTrace1,
                normalizedTrace2,
                unnormalizedTrace2,
                normalizedTrace3,
                unnormalizedTrace3,
                trace4,
                trace5
            ];

            var layout = {
                showlegend: false,
                xaxis: {
                    visible: false,
                    zeroline: false,
                    domain: [0.25, 1],
                    tickvals: vertXAxis,
                    ticktext: labels,
                    side: 'top'
                },

                yaxis: {
                    visible: false,
                    zeroline: false,
                    domain: [0, 0.75],
                    tickvals: horizYAxis,
                    ticktext: labels
                },

                hovermode: 'closest',

                xaxis2: {
                    domain: [0, 0.2],
                    autorange: 'reversed'
                },
                yaxis2: {
                    domain: [0, 0.7]
                },

                xaxis3: {
                    domain: [0, 0.1],
                    zeroline: false,
                    showgrid: false,
                    autorange: 'reversed',
                    visible: false
                },
                yaxis3: {
                    domain: [0.95, 1],
                    zeroline: false,
                    showgrid: false,
                    visible: false
                },

                xaxis4: {
                    domain: [0.15, 1]
                },
                yaxis4: {
                    domain: [0.8, 0.95]
                },

                updatemenus: [{
                    x: 0.75,
                    y: 1.25,
                    buttons: [{
                            method: 'restyle',
                            args: ['visible', [false, true, false, true, false, true, true, true]],
                            label: 'Confusion Matrix'
                        },
                        {
                            method: 'restyle',
                            args: ['visible', [true, false, true, false, true, false, true, true]],
                            label: 'Normalized Confusion Matrix'
                        }
                    ]
                }],

                // axis labels
                annotations: [{
                        xref: 'paper',
                        yref: 'paper',
                        x: -0.01,
                        xanchor: 'right',
                        y: 0.5,
                        yanchor: 'middle',
                        text: 'True Label',
                        showarrow: false,
                        textangle: 270,
                        font: {
                            size: axisLabelSize
                        }
                    },

                    {
                        xref: 'paper',
                        yref: 'paper',
                        x: 0.5,
                        xanchor: 'center',
                        y: 1.07,
                        yanchor: 'top',
                        text: 'Predicted Label',
                        showarrow: false,
                        font: {
                            size: axisLabelSize
                        }
                    }
                ]
            };
            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };
            Plotly.react(elementRef, data, layout, config);
        },

        barChart: function PlotBarChart(elementRef, chartData, mode) {
            var config = {
                displayModeBar: false,
                displaylogo: false
            };

            var data = chartData;

            var chartMargin = {
                l: 40,
                r: 0,
                b: 30,
                t: 40,
                pad: 4
            }
            Plotly.react(elementRef, data, {
                autosize: true,
                margin: chartMargin,
                barmode: mode,
                showlegend: false,
                yaxis: {
                    showticklabels: false,
                    showgrid: false,
                }
            }, config);
        },
        horizontalBarChart: function PlotHorizontalBarChart(elementRef, list, mode) {
            var config = {
                displayModeBar: false,
                displaylogo: false
            };

            var chartData = []
            for (var i = 0; i < Object.keys(list).length; i++) {
                var key = Object.keys(list)[i];
                var val = Object.values(list)[i].count;
                chartData[i] = {
                    x: [val],
                    y: ['Status'],
                    name: key,
                    orientation: 'h',
                    type: 'bar',
                    text: key + ' (' + val + ')',
                    textposition: 'inside',
                    hoverinfo: 'none',
                    marker: {
                        color: Object.values(list)[i].color,
                        width: 1
                    },
                }
            }

            var chartMargin = {
                l: 0,
                r: 0,
                b: 0,
                t: 0,
                pad: 4
            }

            var layout = {
                autosize: true,
                margin: chartMargin,
                barmode: mode,
                showlegend: false,
                height: 50,
                xaxis: {
                    zeroline: false,
                    showticklabels: false,
                    showgrid: false,
                    tickangle: -45
                }
            };
            Plotly.react(elementRef, chartData, layout, config);
        },
        heatmapBarChart: function PlotHeatmapBarChart(elementRef, chartData) {
            var config = {
                displayModeBar: false,
                displaylogo: false
            };

            var gray = 20;
            var red = 50;
            var blue = 100;
            var scaleMap = {
                CompletedWorst: 0,
                CompletedBest: 0.1,
                NotStarted: 0.2,
                Failed: 0.5,
                Running: 1
            }
            var colorScaleValue = [
                [scaleMap.CompletedWorst, chartData.colors.lightgreen],
                [scaleMap.CompletedBest, chartData.colors.green],
                [scaleMap.NotStarted, chartData.colors.gray],
                [scaleMap.Failed, chartData.colors.red],
                [scaleMap.Running, chartData.colors.blue]
            ];

            var runs = _.sortBy(chartData.child_runs, 'run_number');
            var metrics = _.get(chartData, 'child_runs_metrics');
            var isMax = _.endsWith(_.get(metrics, 'series[1].name'), '_max');
            var categories = _.get(metrics, 'series[0].categories', []);
            var primaryMetrics = _.get(metrics, 'series[0].data', []);
            var metrics_mapping = _.zipObject(categories, primaryMetrics);
            var max = _.max(primaryMetrics);
            var min = _.min(primaryMetrics);

            var data = _.reduce(runs, function (result, run) {
                var iteration = run.iteration;
                var status = run.status;
                var metric = _.get(metrics_mapping, iteration);
                var color = gray;
                var text = iteration + ', ' + status;

                if (status === 'NotStarted') {
                    color = gray;
                } else if (status === 'Failed' || (status === 'Completed' && (_.isNaN(metric) || metric === 'NaN'))) {
                    color = red;
                } else if (status === 'Completed') {
                    if (max === min) {
                        color = 10;
                    } else if (isMax) {
                        color = 10 * (metric - min) / (max - min);
                    } else {
                        color = 10 * (max - metric) / (max - min);
                    }
                } else {
                    color = blue;
                }
                
                result.texts[0].push(text);
                result.xValues.push(iteration);
                result.zValues[0].push(color);

                return result;
            }, {
                texts: [[]],
                xValues: [],
                yValues: ['Status'],
                zValues: [[]]
            });

            var hm_data = {
                x: data.xValues,
                y: data.yValues,
                z: data.zValues,
                text: data.texts,
                hoverinfo: 'text',
                zmin: 0,
                zmax: 100,
                xgap: 3,
                ygap: 3,
                type: 'heatmap',
                colorscale: colorScaleValue,
                showscale: false,
                zauto: false,
                zsmooth: false,
                autocolorscale: false
            };

            var chartMargin = {
                l: 50,
                r: 0,
                b: 50,
                t: 10,
                pad: 4
            }

            var dtick = categories.length <= 10 ? 1 : Math.floor(categories.length / 10)
            if (dtick > 10) {
                dtick = _.floor(dtick, -1); // round down to 10
            } else if (dtick > 5) {
                dtick = 10;
            } else if (dtick > 2) {
                dtick = 5;
            }

            var layout = {
                xaxis: {
                    fixedrange: false,
                    autorange: true,
                    showgrid: false,
                    zeroline: false,
                    showticklabels: true,
                    type: 'category',
                    dtick: dtick
                },
                height: 100,
                margin: chartMargin,
            };

            Plotly.react(elementRef, [hm_data], layout, config);
        },
        changeableLineChart: function PlotChangeableLineChart(elementRef, dropElementRef, chartData, showLegend) {
            function getMetricData(chosenMetric) {
                var currentMetric;
                _.forOwn(chartData.series, function (metric) {
                    if (metric[0].name === chosenMetric) {
                        currentMetric = metric;
                    }
                })
                return currentMetric;
            }

            //Configurations for plotting chart
            function setBubblePlot(chosenMetric) {
                var metric = getMetricData(chosenMetric);
                var trace1 = {
                    x: metric[0].categories,
                    y: metric[0].data,
                    mode: metric[0].mode,
                    marker: {
                        size: 12,
                        opacity: 0.8
                    },
                    name: metric[0].name
                };
                var trace2 = {
                    x: metric[1].categories,
                    y: metric[1].data,
                    mode: metric[1].mode,
                    line: {
                        width: 2,
                        shape: 'hv'
                    },
                    name: metric[1].name
                };

                var data = [trace1, trace2];

                var chartMargin = {
                    l: 40,
                    r: 0,
                    b: 40,
                    t: 80,
                    pad: 4
                }

                var layout = {
                    autosize: true,
                    title: 'AutoML Run with metric : ' + chosenMetric,
                    margin: chartMargin,
                    showlegend: showLegend,
                    type: 'category',
                }

                var config = {
                    modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                    displaylogo: false
                };

                Plotly.react(elementRef, data, layout, config);
            };

            function createOrUpdateDropdown(primaryMetricName) {
                var $sel = $(dropElementRef).find('#pm_dropdown');
                if ($sel.length < 1) {
                    $sel = $('<select id="pm_dropdown"></select>').appendTo(dropElementRef);
                }

                $sel.find('option').remove().end();
                _.forOwn(chartData.series, function (metric) {
                    var metric_obj = _.first(metric);
                    if (_.isEmpty(metric_obj) || !metric_obj.name || !_.isNumber(_.first(metric_obj.data))) {
                        return;
                    }

                    $('<option>').text(metric_obj.name).appendTo($sel);
                });

                // set primary metric
                $sel.val(primaryMetricName);
                $sel.off('change').on('change', function () {
                    setBubblePlot($(this).val());
                });
            }

            // Default plot to primary metric
            createOrUpdateDropdown(chartData.primaryMetricName);
            setBubblePlot(chartData.primaryMetricName);

        },
        predictedVsTrueChart: function PredictedVsTrueChart(elementRef, predictions_metrics) {
            var true_vals_bins_start = predictions_metrics.true_vals_bins_start;
            var true_vals_bins_end = predictions_metrics.true_vals_bins_end;
            var avg_pred_vals_per_bin = predictions_metrics.avg_pred_vals_per_bin;
            var pred_error_per_bin = predictions_metrics.pred_error_per_bin;
            var true_vals_count_per_bin = predictions_metrics.true_vals_count_per_bin;

            var true_vals_x = _.zipWith(true_vals_bins_start, true_vals_bins_end, function (
                start,
                end
            ) {
                return (start + end) / 2;
            });

            // lower band
            var pred_err_lower_vals = _.zipWith(
                avg_pred_vals_per_bin,
                pred_error_per_bin,
                function (avg, err) {
                    return avg - err;
                }
            );
            var pred_err_lower_data = {
                x: true_vals_x,
                y: pred_err_lower_vals,
                showlegend: false,
                line: {
                    width: 0,
                    color: "rgba(0,176,246,0.2)"
                },
                marker: {
                    color: "rgba(0,176,246,0.2)"
                },
                mode: "lines",
                name: "Lower Bound",
                type: "scatter"
            };

            // actual values
            var avg_pred_vals_data = {
                x: true_vals_x,
                y: avg_pred_vals_per_bin,
                fill: "tonexty",
                fillcolor: "rgba(0,176,246,0.2)",
                line: {
                    color: "rgb(0,176,246)"
                },
                mode: "lines",
                name: "Actual",
                type: "scatter"
            };

            // upper band
            var pred_err_upper_vals = _.zipWith(
                avg_pred_vals_per_bin,
                pred_error_per_bin,
                function (avg, err) {
                    return avg + err;
                }
            );
            var pred_err_upper_data = {
                x: true_vals_x,
                y: pred_err_upper_vals,
                showlegend: false,
                fill: "tonexty",
                fillcolor: "rgba(0,176,246,0.2)",
                line: {
                    width: 0,
                    color: "rgba(0,176,246,0.2)"
                },
                marker: {
                    color: "rgba(0,176,246,0.2)"
                },
                mode: "lines",
                name: "Upper Bound",
                type: "scatter"
            };

            var true_val_start = _.first(true_vals_x);
            var true_val_end = _.last(true_vals_x);
            var true_vals = [true_val_start, true_val_end];
            var true_vals_data = {
                x: true_vals,
                y: true_vals,
                name: "Ideal",
                mode: "lines",
                line: {
                    dash: "dot",
                    width: 4,
                    color: "rgb(49, 150, 63)"
                }
            };
            var true_vals_cnt_data = {
                x: true_vals_x,
                y: true_vals_count_per_bin,
                hoverinfo: "y",
                showlegend: false,
                xaxis: "x",
                yaxis: "y2",
                type: "bar",
                marker: {
                    color: "rgb(36, 84, 160)"
                }
            };
            var data = [
                pred_err_lower_data,
                avg_pred_vals_data,
                pred_err_upper_data,
                true_vals_data,
                true_vals_cnt_data
            ];
            var layout = {
                title: "Predicted vs. Actual",
                xaxis: {
                    title: "True Value",
                    showline: true,
                    zeroline: false
                },
                yaxis: {
                    title: "Predicted Value",
                    domain: [0.4, 1]
                },
                yaxis2: {
                    title: "Bin Count",
                    domain: [0, 0.2]
                }
            };

            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };
            Plotly.react(elementRef, data, layout, config);
        },
        twoDimensionSurfaceChart: function PlotTwoDimensionSurfaceChart(elementRef, chartData, heatMapOrder) {
            if (!chartData || !chartData.child_runs || chartData.allArguments.length === 0) return;

            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };

            var allArguments = chartData.allArguments;
            var child_runs = chartData.child_runs;
            var metricName = chartData.metricName;
            var paraX = allArguments[0];
            var paraY = allArguments.length > 1 ? allArguments[1] : allArguments[0];
            allArguments.push(chartData.metricName);

            var data = [{
                x: child_runs.map(data => data[_param(paraX, metricName)]),
                y: child_runs.map(data => data[_param(paraY, metricName)]),
                text: child_runs.map(data => allArguments.map(para => para + ': ' + data[_param(para, metricName)]).join('<br>')),
                hoverinfo: 'text',
                mode: 'markers',
                marker: {
                    size: 6,
                    color: child_runs.map(data => data.best_metric),
                    colorscale: 'Viridis',
                    opacity: 0.8,
                    showscale: true,
                    reversescale: heatMapOrder === 'desc' ? false : (heatMapOrder === 'asc' ? true : undefined),
                    colorbar: {
                        len: 0.6,
                        x: 1,
                        xanchor: 'left',
                        yanchor: 'bottom',
                        y: 0.05,
                        ypad: 5,
                        title: 'Metric - ' + chartData.metricName,
                    }
                },
                type: 'scatter'
            }];

            var chartMargin = {
                l: 40,
                r: 0,
                b: 40,
                t: 40,
                pad: 4
            }

            var layout = {
                autosize: true,
                margin: chartMargin,
                xaxis: {
                    title: paraX,
                },
                yaxis: {
                    title: paraY,
                },
                updatemenus: [{
                    x: 1,
                    y: 0.95,
                    xanchor: 'left',
                    yanchor: 'top',
                    showactive: true,
                    active: 0,
                    buttons: allArguments.sort((x, y) => {
                            return x == paraX ? -1 : y == paraX ? 1 : 0;
                        })
                        .map(a => {
                            return {
                                method: 'update',
                                args: [{
                                        'x': [child_runs.map(data => data[_param(a, metricName)])]
                                    },
                                    {
                                        'xaxis.title': a
                                    }
                                ],
                                label: _paramLabel(a, metricName)
                            }
                        })
                }, {
                    x: 1,
                    y: 0.85,
                    xanchor: 'left',
                    yanchor: 'top',
                    buttons: allArguments.sort((x, y) => {
                            return x == paraY ? -1 : y == paraY ? 1 : 0;
                        })
                        .map(a => {
                            return {
                                method: 'update',
                                args: [{
                                        'y': [child_runs.map(data => data[_param(a, metricName)])]
                                    },
                                    {
                                        'yaxis.title': a
                                    }
                                ],
                                label: _paramLabel(a, metricName)
                            }
                        })
                }],
            }

            Plotly.react(elementRef, data, layout, config);
        },
        threeDimensionSurfaceChart: function PlotThreeDimensionSurfaceChart(elementRef, chartData, heatMapOrder) {
            if (!chartData || !chartData.child_runs || chartData.allArguments.length === 0) return;

            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };

            var allArguments = chartData.allArguments;
            var child_runs = chartData.child_runs;
            var metricName = chartData.metricName;
            var paraX = allArguments[0];
            var paraY = allArguments.length > 1 ? allArguments[1] : allArguments[0];
            var paraZ = allArguments.length > 2 ? allArguments[2] : chartData.metricName;
            allArguments.push(chartData.metricName);

            var data = [{
                x: child_runs.map(data => data[_param(paraX, metricName)]),
                y: child_runs.map(data => data[_param(paraY, metricName)]),
                z: child_runs.map(data => data[_param(paraZ, metricName)]),
                text: child_runs.map(data => allArguments.map(para => para + ': ' + data[_param(para, metricName)]).join('<br>')),
                hoverinfo: 'text',
                mode: 'markers',
                marker: {
                    size: 6,
                    color: child_runs.map(data => data.best_metric),
                    colorscale: 'Viridis',
                    opacity: 0.8,
                    showscale: true,
                    reversescale: heatMapOrder === 'desc' ? false : (heatMapOrder === 'asc' ? true : undefined),
                    colorbar: {
                        len: 0.6,
                        x: 1,
                        xanchor: 'left',
                        yanchor: 'bottom',
                        y: 0.05,
                        ypad: 5,
                        title: 'Metric - ' + chartData.metricName,
                    }
                },
                type: 'scatter3d'
            }];

            var chartMargin = {
                l: 0,
                r: 0,
                b: 0,
                t: 0,
                pad: 0
            }

            var layout = {
                autosize: true,
                margin: chartMargin,
                scene: {
                    aspectratio: {
                        x: 0.9,
                        y: 0.9,
                        z: 0.9
                    },
                    xaxis: {
                        title: paraX,
                    },
                    yaxis: {
                        title: paraY,
                    },
                    zaxis: {
                        title: paraZ,
                    }
                },
                updatemenus: [{
                        x: 1,
                        y: 0.95,
                        xanchor: 'left',
                        yanchor: 'top',
                        showactive: true,
                        active: 0,
                        buttons: allArguments.sort((x, y) => {
                                return x == paraX ? -1 : y == paraX ? 1 : 0;
                            })
                            .map(a => {
                                return {
                                    method: 'update',
                                    args: [{
                                            'x': [child_runs.map(data => data[_param(a, metricName)])]
                                        },
                                        {
                                            'scene.xaxis.title': a
                                        }
                                    ],
                                    label: _paramLabel(a, metricName)
                                }
                            })
                    }, {
                        x: 1,
                        y: 0.85,
                        xanchor: 'left',
                        yanchor: 'top',
                        buttons: allArguments.sort((x, y) => {
                                return x == paraY ? -1 : y == paraY ? 1 : 0;
                            })
                            .map(a => {
                                return {
                                    method: 'update',
                                    args: [{
                                            'y': [child_runs.map(data => data[_param(a, metricName)])]
                                        },
                                        {
                                            'scene.yaxis.title': a
                                        }
                                    ],
                                    label: _paramLabel(a, metricName)
                                }
                            })
                    },
                    {
                        x: 1,
                        y: 0.75,
                        xanchor: 'left',
                        yanchor: 'top',
                        buttons: allArguments.sort((x, y) => {
                                return x == paraZ ? -1 : y == paraZ ? 1 : 0;
                            })
                            .map(a => {
                                return {
                                    method: 'update',
                                    args: [{
                                            'z': [child_runs.map(data => data[_param(a, metricName)])]
                                        },
                                        {
                                            'scene.zaxis.title': a
                                        }
                                    ],
                                    label: _paramLabel(a, metricName)
                                }
                            })
                    }
                ],
            }

            Plotly.react(elementRef, data, layout, config);
        },
        parallelCoordinatesChart: function PlotParallelCoordinatesChart(elementRef, chartData, heatMapOrder) {

            var allArguments = chartData.allArguments;
            var child_runs = chartData.child_runs;
            var metricName = chartData.metricName;

            var config = {
                displayModeBar: false,
                displaylogo: false
            };

            var layout = {
                autosize: true
            };

            var trace = {
                type: 'parcoords',
                line: {
                    color: child_runs.map(data => data.best_metric),
                    colorscale: 'Viridis',
                    opacity: 0.8,
                    showscale: true,
                    reversescale: heatMapOrder === 'desc' ? false : (heatMapOrder === 'asc' ? true : undefined),
                    colorbar: {
                        title: chartData.metricName,
                    }
                },
                dimensions: []
            };

            allArguments.forEach(arg => {
                trace.dimensions.push({
                    label: arg,
                    values: child_runs.map(data => data[_param(arg, metricName)])
                });
            });


            var data = [trace]

            Plotly.react(elementRef, data, layout, config);
        },
        pieChart: function PlotPieChart(elementRef, chartData, showLegend){
            var config = {
                modeBarButtonsToRemove: ['toImage', 'sendDataToCloud'],
                displaylogo: false
            };

            var values = [];
            var labels = [];
            var textposition = [];

            for(var key in chartData) {
                if(chartData.hasOwnProperty(key)) {
                    values.push(chartData[key]);
                    labels.push(key.split('NodeCount')[0]);
                    textposition.push(chartData[key]===0 ? 'none' : 'inside')
                }
            }
            
            var data = [{
                values: values,
                labels: labels,
                textinfo: 'label+value',
                textposition: textposition,
                type: 'pie'
              }];
              

            var chartMargin = {
                l: 0,
                r: 0,
                b: 20,
                t: 20,
                pad: 4
            }

            var layout = {
                height: 140,
                autosize: true,
                margin: chartMargin,
                showlegend: false,
                xaxis: {}
            }

            Plotly.react(elementRef, data, layout, config);
        },
        purge: function PlotPurge(elementRef) {
            Plotly.purge(elementRef);
        },

    }

    return chartModule;
});